@@ -52,7 +52,6 @@ static int cmd_add_dev(int argc, char **argv)
 	DIR	*dirstream = NULL;
 	int discard = 1;
 	int force = 0;
-	char estr[100];
 
 	while (1) {
 		int c;
@@ -97,9 +96,8 @@ static int cmd_add_dev(int argc, char **argv)
 		int mixed = 0;
 		char *path;
 
-		res = test_dev_for_mkfs(argv[i], force, estr);
+		res = test_dev_for_mkfs(argv[i], force);
 		if (res) {
-			fprintf(stderr, "%s", estr);
 			ret++;
 			continue;
 		}
@@ -128,6 +126,7 @@ static int cmd_add_dev(int argc, char **argv)
 			goto error_out;
 		}
 
+		memset(&ioctl_args, 0, sizeof(ioctl_args));
 		strncpy_null(ioctl_args.name, path);
 		res = ioctl(fdmnt, BTRFS_IOC_ADD_DEV, &ioctl_args);
 		e = errno;
@@ -177,13 +176,14 @@ static int cmd_rm_dev(int argc, char **argv)
 			ret++;
 			continue;
 		}
+		memset(&arg, 0, sizeof(arg));
 		strncpy_null(arg.name, argv[i]);
 		res = ioctl(fdmnt, BTRFS_IOC_RM_DEV, &arg);
 		e = errno;
 		if (res) {
 			const char *msg;
 
-			if (ret > 0)
+			if (res > 0)
 				msg = btrfs_err_str(res);
 			else
 				msg = strerror(e);
@@ -314,7 +314,8 @@ static int cmd_ready_dev(int argc, char **argv)
 		goto out;
 	}
 
-	strncpy(args.name, path, BTRFS_PATH_NAME_MAX);
+	memset(&args, 0, sizeof(args));
+	strncpy_null(args.name, path);
 	ret = ioctl(fd, BTRFS_IOC_DEVICES_READY, &args);
 	if (ret < 0) {
 		fprintf(stderr, "ERROR: unable to determine if the device '%s'"
@@ -583,8 +584,11 @@ out:
 	return !!ret;
 }
 
+static const char device_cmd_group_info[] =
+"manage and query devices in the filesystem";
+
 const struct cmd_group device_cmd_group = {
-	device_cmd_group_usage, NULL, {
+	device_cmd_group_usage, device_cmd_group_info, {
 		{ "add", cmd_add_dev, cmd_add_dev_usage, NULL, 0 },
 		{ "delete", cmd_rm_dev, cmd_rm_dev_usage, NULL, 0 },
 		{ "scan", cmd_scan_dev, cmd_scan_dev_usage, NULL, 0 },
