@@ -43,6 +43,7 @@
 #include "utils.h"
 
 static u64 index_cnt = 2;
+static int verbose = 1;
 
 struct directory_name_entry {
 	char *dir_name;
@@ -51,7 +52,15 @@ struct directory_name_entry {
 	struct list_head list;
 };
 
-static int make_root_dir(struct btrfs_root *root, int mixed)
+struct mkfs_allocation {
+	u64 data;
+	u64 metadata;
+	u64 mixed;
+	u64 system;
+};
+
+static int make_root_dir(struct btrfs_root *root, int mixed,
+				struct mkfs_allocation *allocation)
 {
 	struct btrfs_trans_handle *trans;
 	struct btrfs_key location;
@@ -68,6 +77,7 @@ static int make_root_dir(struct btrfs_root *root, int mixed)
 				     BTRFS_BLOCK_GROUP_SYSTEM,
 				     BTRFS_FIRST_CHUNK_TREE_OBJECTID,
 				     0, BTRFS_MKFS_SYSTEM_GROUP_SIZE);
+	allocation->system += BTRFS_MKFS_SYSTEM_GROUP_SIZE;
 	BUG_ON(ret);
 
 	if (mixed) {
@@ -87,7 +97,7 @@ static int make_root_dir(struct btrfs_root *root, int mixed)
 					     BTRFS_FIRST_CHUNK_TREE_OBJECTID,
 					     chunk_start, chunk_size);
 		BUG_ON(ret);
-		printf("Created a data/metadata chunk of size %llu\n", chunk_size);
+		allocation->mixed += chunk_size;
 	} else {
 		ret = btrfs_alloc_chunk(trans, root->fs_info->extent_root,
 					&chunk_start, &chunk_size,
@@ -101,6 +111,7 @@ static int make_root_dir(struct btrfs_root *root, int mixed)
 					     BTRFS_BLOCK_GROUP_METADATA,
 					     BTRFS_FIRST_CHUNK_TREE_OBJECTID,
 					     chunk_start, chunk_size);
+		allocation->metadata += chunk_size;
 		BUG_ON(ret);
 	}
 
@@ -122,6 +133,7 @@ static int make_root_dir(struct btrfs_root *root, int mixed)
 					     BTRFS_BLOCK_GROUP_DATA,
 					     BTRFS_FIRST_CHUNK_TREE_OBJECTID,
 					     chunk_start, chunk_size);
+		allocation->data += chunk_size;
 		BUG_ON(ret);
 	}
 
@@ -181,7 +193,9 @@ static void recow_roots(struct btrfs_trans_handle *trans,
 }
 
 static int create_one_raid_group(struct btrfs_trans_handle *trans,
-			      struct btrfs_root *root, u64 type)
+			      struct btrfs_root *root, u64 type,
+			      struct mkfs_allocation *allocation)
+
 {
 	u64 chunk_start;
 	u64 chunk_size;
@@ -197,6 +211,18 @@ static int create_one_raid_group(struct btrfs_trans_handle *trans,
 	ret = btrfs_make_block_group(trans, root->fs_info->extent_root, 0,
 				     type, BTRFS_FIRST_CHUNK_TREE_OBJECTID,
 				     chunk_start, chunk_size);
+	if ((type & BTRFS_BLOCK_GROUP_TYPE_MASK) == BTRFS_BLOCK_GROUP_DATA)
+		allocation->data += chunk_size;
+	else if ((type & BTRFS_BLOCK_GROUP_TYPE_MASK) == BTRFS_BLOCK_GROUP_METADATA)
+		allocation->metadata += chunk_size;
+	else if ((type & BTRFS_BLOCK_GROUP_TYPE_MASK) == BTRFS_BLOCK_GROUP_SYSTEM)
+		allocation->system += chunk_size;
+	else if ((type & BTRFS_BLOCK_GROUP_TYPE_MASK) ==
+			(BTRFS_BLOCK_GROUP_METADATA|BTRFS_BLOCK_GROUP_DATA))
+		allocation->mixed += chunk_size;
+	else
+		BUG_ON(1);
+
 	BUG_ON(ret);
 	return ret;
 }
@@ -204,7 +230,8 @@ static int create_one_raid_group(struct btrfs_trans_handle *trans,
 static int create_raid_groups(struct btrfs_trans_handle *trans,
 			      struct btrfs_root *root, u64 data_profile,
 			      int data_profile_opt, u64 metadata_profile,
-			      int mixed)
+			      int mixed,
+			      struct mkfs_allocation *allocation)
 {
 	u64 num_devices = btrfs_super_num_devices(root->fs_info->super_copy);
 	int ret;
@@ -214,21 +241,21 @@ static int create_raid_groups(struct btrfs_trans_handle *trans,
 
 		ret = create_one_raid_group(trans, root,
 					    BTRFS_BLOCK_GROUP_SYSTEM |
-					    metadata_profile);
+					    metadata_profile, allocation);
 		BUG_ON(ret);
 
 		if (mixed)
 			meta_flags |= BTRFS_BLOCK_GROUP_DATA;
 
 		ret = create_one_raid_group(trans, root, meta_flags |
-					    metadata_profile);
+					    metadata_profile, allocation);
 		BUG_ON(ret);
 
 	}
 	if (!mixed && num_devices > 1 && data_profile) {
 		ret = create_one_raid_group(trans, root,
 					    BTRFS_BLOCK_GROUP_DATA |
-					    data_profile);
+					    data_profile, allocation);
 		BUG_ON(ret);
 	}
 	recow_roots(trans, root);
@@ -263,9 +290,7 @@ static int create_data_reloc_tree(struct btrfs_trans_handle *trans,
 	return 0;
 }
 
-
-static void print_usage(void) __attribute__((noreturn));
-static void print_usage(void)
+static void print_usage(int ret)
 {
 	fprintf(stderr, "usage: mkfs.btrfs [options] dev [ dev ... ]\n");
 	fprintf(stderr, "options:\n");
@@ -281,11 +306,12 @@ static void print_usage(void)
 	fprintf(stderr, "\t-s|--sectorsize SIZE    min block allocation (may not mountable by current kernel)\n");
 	fprintf(stderr, "\t-r|--rootdir DIR        the source directory\n");
 	fprintf(stderr, "\t-K|--nodiscard          do not perform whole device TRIM\n");
-	fprintf(stderr, "\t-O|--features LIST      comma separated list of filesystem features\n");
+	fprintf(stderr, "\t-O|--features LIST      comma separated list of filesystem features, use '-O list-all' to list features\n");
 	fprintf(stderr, "\t-U|--uuid UUID          specify the filesystem UUID\n");
+	fprintf(stderr, "\t-q|--quiet              no messages except errors\n");
 	fprintf(stderr, "\t-V|--version            print the mkfs.btrfs version and exit\n");
 	fprintf(stderr, "%s\n", PACKAGE_STRING);
-	exit(1);
+	exit(ret);
 }
 
 static void print_version(void) __attribute__((noreturn));
@@ -299,21 +325,20 @@ static u64 parse_profile(char *s)
 {
 	if (strcmp(s, "raid0") == 0) {
 		return BTRFS_BLOCK_GROUP_RAID0;
-	} else if (strcmp(s, "raid1") == 0) {
+	} else if (strcasecmp(s, "raid1") == 0) {
 		return BTRFS_BLOCK_GROUP_RAID1;
-	} else if (strcmp(s, "raid5") == 0) {
+	} else if (strcasecmp(s, "raid5") == 0) {
 		return BTRFS_BLOCK_GROUP_RAID5;
-	} else if (strcmp(s, "raid6") == 0) {
+	} else if (strcasecmp(s, "raid6") == 0) {
 		return BTRFS_BLOCK_GROUP_RAID6;
-	} else if (strcmp(s, "raid10") == 0) {
+	} else if (strcasecmp(s, "raid10") == 0) {
 		return BTRFS_BLOCK_GROUP_RAID10;
-	} else if (strcmp(s, "dup") == 0) {
+	} else if (strcasecmp(s, "dup") == 0) {
 		return BTRFS_BLOCK_GROUP_DUP;
-	} else if (strcmp(s, "single") == 0) {
+	} else if (strcasecmp(s, "single") == 0) {
 		return 0;
 	} else {
 		fprintf(stderr, "Unknown profile %s\n", s);
-		print_usage();
 	}
 	/* not reached */
 	return 0;
@@ -891,7 +916,8 @@ static int open_target(char *output_name)
 
 static int create_chunks(struct btrfs_trans_handle *trans,
 			 struct btrfs_root *root, u64 num_of_meta_chunks,
-			 u64 size_of_data)
+			 u64 size_of_data,
+			 struct mkfs_allocation *allocation)
 {
 	u64 chunk_start;
 	u64 chunk_size;
@@ -908,6 +934,7 @@ static int create_chunks(struct btrfs_trans_handle *trans,
 		ret = btrfs_make_block_group(trans, root->fs_info->extent_root, 0,
 					     meta_type, BTRFS_FIRST_CHUNK_TREE_OBJECTID,
 					     chunk_start, chunk_size);
+		allocation->metadata += chunk_size;
 		BUG_ON(ret);
 		set_extent_dirty(&root->fs_info->free_space_cache,
 				 chunk_start, chunk_start + chunk_size - 1, 0);
@@ -922,6 +949,7 @@ static int create_chunks(struct btrfs_trans_handle *trans,
 	ret = btrfs_make_block_group(trans, root->fs_info->extent_root, 0,
 				     data_type, BTRFS_FIRST_CHUNK_TREE_OBJECTID,
 				     chunk_start, size_of_data);
+	allocation->data += size_of_data;
 	BUG_ON(ret);
 	set_extent_dirty(&root->fs_info->free_space_cache,
 			 chunk_start, chunk_start + size_of_data - 1, 0);
@@ -955,7 +983,8 @@ static int make_image(char *source_dir, struct btrfs_root *root, int out_fd)
 	}
 	btrfs_commit_transaction(trans, root);
 
-	printf("Making image is completed.\n");
+	if (verbose)
+		printf("Making image is completed.\n");
 	return 0;
 fail:
 	while (!list_empty(&dir_head.list)) {
@@ -1100,13 +1129,43 @@ static int is_ssd(const char *file)
 	return !atoi((const char *)&rotational);
 }
 
+static void list_all_devices(struct btrfs_root *root)
+{
+	struct btrfs_fs_devices *fs_devices;
+	struct btrfs_device *device;
+	int number_of_devices = 0;
+	u64 total_block_count = 0;
+
+	fs_devices = root->fs_info->fs_devices;
+
+	list_for_each_entry(device, &fs_devices->devices, dev_list)
+		number_of_devices++;
+
+	printf("Number of devices:  %d\n", number_of_devices);
+	/* printf("Total devices size: %10s\n", */
+		/* pretty_size(total_block_count)); */
+	printf("Devices:\n");
+	printf("   ID        SIZE  PATH\n");
+	list_for_each_entry_reverse(device, &fs_devices->devices, dev_list) {
+		char dev_uuid[BTRFS_UUID_UNPARSED_SIZE];
+
+		uuid_unparse(device->uuid, dev_uuid);
+		printf("  %3llu  %10s  %s\n",
+			device->devid,
+			pretty_size(device->total_bytes),
+			device->name);
+		total_block_count += device->total_bytes;
+	}
+
+	printf("\n");
+}
+
 int main(int ac, char **av)
 {
 	char *file;
 	struct btrfs_root *root;
 	struct btrfs_trans_handle *trans;
 	char *label = NULL;
-	char *first_file;
 	u64 block_count = 0;
 	u64 dev_block_count = 0;
 	u64 blocks[7];
@@ -1136,9 +1195,9 @@ int main(int ac, char **av)
 	u64 source_dir_size = 0;
 	int dev_cnt = 0;
 	int saved_optind;
-	char estr[100];
-	char *fs_uuid = NULL;
+	char fs_uuid[BTRFS_UUID_UNPARSED_SIZE] = { 0 };
 	u64 features = BTRFS_MKFS_DEFAULT_FEATURES;
+	struct mkfs_allocation allocation = { 0 };
 
 	while(1) {
 		int c;
@@ -1158,10 +1217,12 @@ int main(int ac, char **av)
 			{ "nodiscard", no_argument, NULL, 'K' },
 			{ "features", required_argument, NULL, 'O' },
 			{ "uuid", required_argument, NULL, 'U' },
+			{ "quiet", 0, NULL, 'q' },
+			{ "help", no_argument, NULL, GETOPT_VAL_HELP },
 			{ NULL, 0, NULL, 0}
 		};
 
-		c = getopt_long(ac, av, "A:b:fl:n:s:m:d:L:O:r:U:VMK",
+		c = getopt_long(ac, av, "A:b:fl:n:s:m:d:L:O:r:U:VMKq",
 				long_options, NULL);
 		if (c < 0)
 			break;
@@ -1217,11 +1278,8 @@ int main(int ac, char **av)
 				break;
 			case 'b':
 				block_count = parse_size(optarg);
-				if (block_count <= BTRFS_MKFS_SMALL_VOLUME_SIZE) {
-					fprintf(stdout,
-				"SMALL VOLUME: forcing mixed metadata/data groups\n");
+				if (block_count <= BTRFS_MKFS_SMALL_VOLUME_SIZE)
 					mixed = 1;
-				}
 				zero_end = 0;
 				break;
 			case 'V':
@@ -1232,13 +1290,18 @@ int main(int ac, char **av)
 				source_dir_set = 1;
 				break;
 			case 'U':
-				fs_uuid = optarg;
+				strncpy(fs_uuid, optarg,
+					BTRFS_UUID_UNPARSED_SIZE - 1);
 				break;
 			case 'K':
 				discard = 0;
 				break;
+			case 'q':
+				verbose = 0;
+				break;
+			case GETOPT_VAL_HELP:
 			default:
-				print_usage();
+				print_usage(c != GETOPT_VAL_HELP);
 		}
 	}
 	sectorsize = max(sectorsize, (u32)sysconf(_SC_PAGESIZE));
@@ -1247,7 +1310,7 @@ int main(int ac, char **av)
 	saved_optind = optind;
 	dev_cnt = ac - optind;
 	if (dev_cnt == 0)
-		print_usage();
+		print_usage(1);
 
 	if (source_dir_set && dev_cnt > 1) {
 		fprintf(stderr,
@@ -1255,7 +1318,7 @@ int main(int ac, char **av)
 		exit(1);
 	}
 
-	if (fs_uuid) {
+	if (*fs_uuid) {
 		uuid_t dummy_uuid;
 
 		if (uuid_parse(fs_uuid, dummy_uuid) != 0) {
@@ -1271,10 +1334,8 @@ int main(int ac, char **av)
 	while (dev_cnt-- > 0) {
 		file = av[optind++];
 		if (is_block_device(file))
-			if (test_dev_for_mkfs(file, force_overwrite, estr)) {
-				fprintf(stderr, "Error: %s", estr);
+			if (test_dev_for_mkfs(file, force_overwrite))
 				exit(1);
-			}
 	}
 
 	optind = saved_optind;
@@ -1283,8 +1344,9 @@ int main(int ac, char **av)
 	file = av[optind++];
 	ssd = is_ssd(file);
 
-	if (is_vol_small(file)) {
-		printf("SMALL VOLUME: forcing mixed metadata/data groups\n");
+	if (is_vol_small(file) || mixed) {
+		if (verbose)
+			printf("SMALL VOLUME: forcing mixed metadata/data groups\n");
 		mixed = 1;
 	}
 
@@ -1294,7 +1356,7 @@ int main(int ac, char **av)
 	*/
 	if (!mixed) {
 		if (!metadata_profile_opt) {
-			if (dev_cnt == 1 && ssd)
+			if (dev_cnt == 1 && ssd && verbose)
 				printf("Detected a SSD, turning off metadata "
 				"duplication.  Mkfs with -m dup if you want to "
 				"force metadata duplication.\n");
@@ -1361,15 +1423,15 @@ int main(int ac, char **av)
 		}
 	}
 	ret = test_num_disk_vs_raid(metadata_profile, data_profile,
-			dev_cnt, mixed, estr);
-	if (ret) {
-		fprintf(stderr, "Error: %s\n", estr);
+			dev_cnt, mixed);
+	if (ret)
 		exit(1);
-	}
 
 	/* if we are here that means all devs are good to btrfsify */
-	printf("%s\n", PACKAGE_STRING);
-	printf("See %s for more information.\n\n", PACKAGE_URL);
+	if (verbose) {
+		printf("%s\n", PACKAGE_STRING);
+		printf("See %s for more information.\n\n", PACKAGE_URL);
+	}
 
 	dev_cnt--;
 
@@ -1385,7 +1447,6 @@ int main(int ac, char **av)
 				strerror(errno));
 			exit(1);
 		}
-		first_file = file;
 		ret = btrfs_prepare_device(fd, file, zero_end, &dev_block_count,
 					   block_count, &mixed, discard);
 		if (ret) {
@@ -1403,7 +1464,6 @@ int main(int ac, char **av)
 			exit(1);
 		}
 
-		first_file = file;
 		source_dir_size = size_sourcedir(source_dir, sectorsize,
 					     &num_of_meta_chunks, &size_of_data);
 		if(block_count < source_dir_size)
@@ -1429,6 +1489,12 @@ int main(int ac, char **av)
 			nodesize * i;
 	}
 
+	if (group_profile_max_safe_loss(metadata_profile) <
+		group_profile_max_safe_loss(data_profile)){
+		fprintf(stderr,
+			"WARNING: metatdata has lower redundancy than data!\n\n");
+	}
+
 	/*
 	 * FS features that can be set by other means than -O
 	 * just set the bit here
@@ -1441,8 +1507,6 @@ int main(int ac, char **av)
 		features |= BTRFS_FEATURE_INCOMPAT_RAID56;
 	}
 
-	btrfs_process_fs_features(features);
-
 	ret = make_btrfs(fd, file, label, fs_uuid, blocks, dev_block_count,
 			 nodesize, sectorsize, stripesize, features);
 	if (ret) {
@@ -1458,7 +1522,7 @@ int main(int ac, char **av)
 	}
 	root->fs_info->alloc_start = alloc_start;
 
-	ret = make_root_dir(root, mixed);
+	ret = make_root_dir(root, mixed, &allocation);
 	if (ret) {
 		fprintf(stderr, "failed to setup the root directory\n");
 		exit(1);
@@ -1466,7 +1530,8 @@ int main(int ac, char **av)
 
 	trans = btrfs_start_transaction(root, 1);
 
-	btrfs_register_one_device(file);
+	if (is_block_device(file))
+		btrfs_register_one_device(file);
 
 	if (dev_cnt == 0)
 		goto raid_groups;
@@ -1506,31 +1571,37 @@ int main(int ac, char **av)
 		ret = btrfs_add_to_fsid(trans, root, fd, file, dev_block_count,
 					sectorsize, sectorsize, sectorsize);
 		BUG_ON(ret);
-		btrfs_register_one_device(file);
+		if (verbose >= 2) {
+			struct btrfs_device *device;
+
+			device = container_of(root->fs_info->fs_devices->devices.next,
+					struct btrfs_device, dev_list);
+			printf("adding device %s id %llu\n", file,
+				(unsigned long long)device->devid);
+		}
+
+		if (is_block_device(file))
+			btrfs_register_one_device(file);
 	}
 
 raid_groups:
 	if (!source_dir_set) {
 		ret = create_raid_groups(trans, root, data_profile,
 				 data_profile_opt, metadata_profile,
-				 mixed);
+				 mixed, &allocation);
 		BUG_ON(ret);
 	}
 
 	ret = create_data_reloc_tree(trans, root);
 	BUG_ON(ret);
 
-	printf("fs created label %s on %s\n\tnodesize %u leafsize %u "
-	    "sectorsize %u size %s\n",
-	    label, first_file, nodesize, nodesize, sectorsize,
-	    pretty_size(btrfs_super_total_bytes(root->fs_info->super_copy)));
-
 	btrfs_commit_transaction(trans, root);
 
 	if (source_dir_set) {
 		trans = btrfs_start_transaction(root, 1);
 		ret = create_chunks(trans, root,
-				    num_of_meta_chunks, size_of_data);
+				    num_of_meta_chunks, size_of_data,
+				    &allocation);
 		BUG_ON(ret);
 		btrfs_commit_transaction(trans, root);
 
@@ -1538,6 +1609,39 @@ raid_groups:
 		BUG_ON(ret);
 	}
 
+	if (verbose) {
+		char features_buf[64];
+
+		printf("Label:              %s\n", label);
+		printf("UUID:               %s\n", fs_uuid);
+		printf("Node size:          %u\n", nodesize);
+		printf("Sector size:        %u\n", sectorsize);
+		printf("Filesystem size:    %s\n",
+			pretty_size(btrfs_super_total_bytes(root->fs_info->super_copy)));
+		printf("Block group profiles:\n");
+		if (allocation.data)
+			printf("  Data:             %-8s %16s\n",
+				btrfs_group_profile_str(data_profile),
+				pretty_size(allocation.data));
+		if (allocation.metadata)
+			printf("  Metadata:         %-8s %16s\n",
+				btrfs_group_profile_str(metadata_profile),
+				pretty_size(allocation.metadata));
+		if (allocation.mixed)
+			printf("  Data+Metadata:    %-8s %16s\n",
+				btrfs_group_profile_str(data_profile),
+				pretty_size(allocation.mixed));
+		printf("  System:           %-8s %16s\n",
+			btrfs_group_profile_str(metadata_profile),
+			pretty_size(allocation.system));
+		printf("SSD detected:       %s\n", ssd ? "yes" : "no");
+		btrfs_parse_features_to_string(features_buf, features);
+		printf("Incompat features:  %s", features_buf);
+		printf("\n");
+
+		list_all_devices(root);
+	}
+
 	ret = close_ctree(root);
 	BUG_ON(ret);
 	free(label);
